package dk.skat.akfa.transformation.javacallout;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Vector;

/**
 * <p>Implements AKFA message structure. Input is read from a filestream and written to a filestream</p>
 * <p><pre>
 * Filnavn:  Maskinelt a-kasse og fagligt kontingent
 * Individnavn:  A-kasse / fagforening
 * Filnr.:   285
 * Individnr:  2260
 * Individstrrelse 230 (min/max tegn)
 * Start. pos Lngde bytes samt repr. Dataforkortelse  Bemrkninger
 * 1            4 (Numerisk)          INDIVIDNR         K: 2260.
 * 5            8 (Numerisk)          SENR              Udfyldes med det SE-nr., den oplysningspligtige er tildelt af SKAT.
 * 13          10 (Numerisk)          CPRNR/SENR        Udfyldes med det CPR-nr./SE-nr., der er tilknyttet den enkelte indberetning. SE-nr. udfyldes med to foranstillede nuller.
 * 23          15 (Alfanumerisk)      MEDLEMSNR.        Kan udfyldes med den enkeltes medlemsnummer.
 * 38           2 (Numerisk)          KALENDERR        Udfyldes med de to sidste cifre i det kalenderr indberetningen vedrrer.
 * 40           1 (Alfanumerisk)      FORTEGN           Udfyldes med blank, hvis positiv, eller - (minus), hvis negativ.
 * 41           5 (Numerisk)          FRADRAGSBERET. INDBETALINGER  Udfyldes med det fradragsberettigede belb i hele kroner.
 * 46           1 (Numerisk)          BELBSART         Udfyldes med en af flgende koder:
 *  1 = A-kassebidrag
 *  2 = kontingent til en faglig forening
 *  3 = klubkontingent
 *  4 = fagforenings- og klubkontingent
 *  5 = faglig forening, der anvender kontingentet i en ikke-erhvervsmssig del og en erhvervsmssig del
 *  6 = efterlnskontingent
 *  7 = flekspensionsbidrag
 *  8 = Privat arbejdslshedsforsikring.
 *
 * 47           1 (Alfanumerisk)      FORTEGN           Udfyldes med blank, hvis positiv, eller - (minus), hvis negativ.
 * 48           5 (Numerisk)          ERHVERVSMSSIG INDBETALING Udfyldes med den del af kontingentet, der kan henfres til erhvervsmssig indbetaling i hele kroner.
 * 53           1 (Alfanumerisk)      BETALINGSKODE     Udfyldes med vrdien = 1, sfremt indberetningen vedrrer kontingent, der skal opdeles i en ikke-erhvervsmssig del og en erhvervsmssig del, og arbejdsgiveren har betalt kontingentet. Ellers udffyldes feltet med 0/blank.
 * 54           1 (Alfanumerisk)      GRUPPELIVSFORS.   Udfyldes med vrdien = 1, sfremt en del af kontingentet anvendes til gruppelivsforsikring. Sfremt prmie til gruppelivsforsikring opkrves srskilt, skal koden ikke opsttes, og prmiebelb skal ikke medtages i belbsoplysningen. Ellers udffyldes feltet med 0/blank.
 * 55           1 (Alfanumerisk)      RETTEKODE         Udfyldes med 0/blank, sfremt indberetningen vedrrer ordinr indberetning (herunder ogs omlevering af et helt bnd efter aftale med SKAT).
 * 1 = korrektion til tidligere indberetning.
 * 2 = rettelse til en tidligere indberetning, der er afvist p fejlliste fra SKAT.
 * 56          40 (Alfanumerisk)      NAVN/ADRESSE 1    Kan udfyldes med medlemmets navn.
 * 96          40 (Alfanumerisk)      NAVN/ADRESSE 2    Kan udfyldes med medlemmets adresse.
 * 136         40 (Alfanumerisk)      NAVN/ADRESSE 3    Kan udfyldes med medlemmets postdistrikt.
 * 176         40 (Alfanumerisk)      NAVN/ADRESSE 4    Kan udfyldes med medlemmets postnummer.
 * 216         15 (Alfanumerisk)      FILLER   -
 * </pre></p>
 * @author Ren Hjortskov Nielsen, SKAT
 * @version 0.1
 */
public class AKFAXMLMessage extends AKFAXMLMessageBase implements AKFAXMLMessageInterface
{
  private byte[] payload;

	public AKFAXMLMessage(boolean debug) {
		this.debug = debug;
	}

	/**
	 * Produce from <p>2260400141362202551605medlemsnr01---x99+11111a-22222bgrnavnadresse1---------------------------xnavnadresse2---------------------------xnavnadresse3---------------------------xnavnadresse4---------------------------x--------------x</p>
   *
   * </p>
	 */
  public void readFromFile(FileInputStream fis, FileOutputStream fos)
		throws Exception
	{
		//if (debug) System.out.println("+readFromFile");
		// Reset message
    if (message==null) { message = new Vector(); }

    boolean empty = message.isEmpty();
    if (empty) {
      message.addElement(getBytesFromString("<fase1:IndberetningEnhed>"));
      message.addElement(getBytesFromString("<fase1:VirksomhedAngivelseDetail>"));
      message.addElement(getBytesFromString("<fase1:AngiverVirksomhedSENummer>"));
      message.addElement(getBytesFromString("<fase1:VirksomhedSENummer>"));

      message.addElement(getBytesFromString(""));

      message.addElement(getBytesFromString("</fase1:VirksomhedSENummer>"));
      message.addElement(getBytesFromString("</fase1:AngiverVirksomhedSENummer>"));
      message.addElement(getBytesFromString("</fase1:VirksomhedAngivelseDetail>"));
    
      message.addElement(getBytesFromString("<fase1:AngivelseInddataDetail>"));
      message.addElement(getBytesFromString("<fase1:AngivelseValgtPeriode>"));
      message.addElement(getBytesFromString("<fase1:AngivelseValgtPeriodeSlutDato>"));
      message.addElement(getBytesFromString("20")); // <= Provided by NTSE AKFA solution
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("-12-31+01:00")); // MM-DD <= Provided by NTSE AKFA solution
      message.addElement(getBytesFromString("</fase1:AngivelseValgtPeriodeSlutDato>"));
      message.addElement(getBytesFromString("</fase1:AngivelseValgtPeriode>"));

      message.addElement(getBytesFromString("<fase1:AKFAAngivelseInddata>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseMedlemCPRSENummer>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseMedlemCPRSENummer>"));

      message.addElement(getBytesFromString("<fase1:AKFAAngivelseBetalingKode>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseBetalingKode>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseBelbArt>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseBelbArt>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseErhvervsmssigIndbetalingBelb>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseErhvervsmssigIndbetalingBelb>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseFradragsberettigetIndbetalingBelb>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseFradragsberettigetIndbetalingBelb>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseGruppelivForsikringKode>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseGruppelivForsikringKode>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseMedlemNummer>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseMedlemNummer>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseRetteKode>"));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("</fase1:AKFAAngivelseRetteKode>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseMedlemNavn><![CDATA["));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("]]></fase1:AKFAAngivelseMedlemNavn>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseMedlemAdresse><![CDATA["));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("]]></fase1:AKFAAngivelseMedlemAdresse>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseMedlemPostdistrikt><![CDATA["));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("]]></fase1:AKFAAngivelseMedlemPostdistrikt>"));
      message.addElement(getBytesFromString("<fase1:AKFAAngivelseMedlemPostNummer><![CDATA["));
      message.addElement(getBytesFromString(""));
      message.addElement(getBytesFromString("]]></fase1:AKFAAngivelseMedlemPostNummer>"));

      message.addElement(getBytesFromString("</fase1:AKFAAngivelseInddata>"));
      message.addElement(getBytesFromString("</fase1:AngivelseInddataDetail>"));
      message.addElement(getBytesFromString("</fase1:IndberetningEnhed>"));
    } 

		readBytes(fis,0,4,0,null); // INDIVIDNR
    message.setElementAt(readBytes(fis,0,8,0,null),4); // SENR
    byte[] angivelsemedlemssenummer = readBytes(fis,0,10,0,null);
    byte[] medlemsnummer = readBytes(fis,0,15,0,null);
    message.setElementAt(readBytes(fis,0,2,0,null),12); // KALENDERR
    message.setElementAt(angivelsemedlemssenummer,18); // CPRNR/SENR
    byte[] fradragsberettigetindbetalingfortegn = readBytes(fis,0,1,0,null);
    byte[] fradragsberettigetindbetaling = readBytes(fis,0,5,0,null);
    byte[] belobart = readBytes(fis,0,1,0,null);
    byte[] erhvervsmaessigindbetalingfortegn = readBytes(fis,0,1,0,null);
    byte[] erhvervsmaessigindbetaling = readBytes(fis,0,5,0,null);
    byte[] betalingskode = readBytes(fis,0,1,0,null);
    message.setElementAt(betalingskode,21); // BETALINGSKODE
    message.setElementAt(belobart,24); // BELBSART
    message.setElementAt(erhvervsmaessigindbetalingfortegn,27); // FORTEGN ?? <= Provided by NTSE AKFA solution, + is left out
    message.setElementAt(erhvervsmaessigindbetaling,28); // ERHVERVSMSSIG INDBETALING
    message.setElementAt(fradragsberettigetindbetalingfortegn,31); // FORTEGN ?? <= Provided by NTSE AKFA solution, + is left out
    message.setElementAt(fradragsberettigetindbetaling,32); // FRADRAGSBERET. INDBETALINGER
    message.setElementAt(readBytes(fis,0,1,0,null),35); // GRUPPELIVSFORS
    message.setElementAt(medlemsnummer,38); // MEDLEMSNR.
    message.setElementAt(readBytes(fis,0,1,0,null),41); // RETTEKODE
    // trim reduces gzipped xml size, but at the cost of cpu time
    message.setElementAt(new String(readBytes(fis,0,40,0,null),"ISO8859_1").trim().getBytes("UTF-8"),44); // NAVN/ADRESSE 1
    message.setElementAt(new String(readBytes(fis,0,40,0,null),"ISO8859_1").trim().getBytes("UTF-8"),47); // NAVN/ADRESSE 2
    message.setElementAt(new String(readBytes(fis,0,40,0,null),"ISO8859_1").trim().getBytes("UTF-8"),50); // NAVN/ADRESSE 3
    message.setElementAt(new String(readBytes(fis,0,40,0,null),"ISO8859_1").trim().getBytes("UTF-8"),53); // NAVN/ADRESSE 4
    
    readBytes(fis,0,15,0,null); // FILLER

    // Discard LF
    readLinefeed(fis);

		//if (debug) System.out.println("-readFromFile");
  }

}