package dk.skat.akfa.transformation.javacallout;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Vector;
import java.nio.channels.FileChannel;
import java.util.Hashtable;

/**
 * <p>Abstract message class used as a library for message processing
 * and as part of enforcing implementation of shared and individual
 * methods</p>
 *
 * @author Ren Hjortskov Nielsen, SKAT
 * @version 0.1
 */
abstract class AKFAXMLMessageBase
{
  protected byte[] payload      = null;
  protected final int CR = 0x0D;
  protected final int LF = 0x0A;
  protected boolean debug = false;
  public Vector message;// Placeholder for xml message

	protected AKFAXMLMessageBase() {
	}

  // Common SKAT XML Structures and AKFA BNF Structures

	// Common methods

	/**
	 * <p>Read len bytes from stream after skipping bytes</p>
	 *
	 * <p>Note: field_separator 0x3B is not automatically removed</p>
	 *
	 * <p>field_separator is NOT removed</p>
	 *
	 * @param fis FileInputStream
	 * @param skip Number of bytes to skip
	 * @param len Number of bytes to read
	 * @param postskip Bytes to skip in the end of a string exclusive field_separator, which is always skipped
	 * @param fieldCodeExpected bytes denoting which field is supposed to be read. If there is no match an exception is thrown
	 * @return byte[] bytes read
	 */
	protected byte[] readBytes(FileInputStream fis, int skip, int len, int postskip, byte[] fieldCodeExpected)
		throws Exception
	{
	  byte[] bb = new byte[1];
		byte[] b = new byte[len];
		// Match expected field code
	  if (fieldCodeExpected == null || fieldCodeExpected.length==0) {
		  if (skip>0) fis.skip((long)skip);
		}
		if ((fis.read(b,0,len))!=-1) {
		  while (postskip>0) {
		    bb[0] = (byte)fis.read();
		    if (debug) System.out.println("skipping ["+(new String(bb))+"]");
		    postskip--;
		  }
		  if (debug) System.out.println("bytesRead => [" + new String(b)+"]");
			return b;
    } else
        throw new Exception("Error reading bytes from stream ["+b[0]+"]");
	}

	/**
	 * Returns the byte representation of a string
	 *
	 * @param str The string
	 */
	protected byte[] getBytesFromString(String str)
		throws Exception
	{
		return str.getBytes("UTF-8");
	}

  /**
   * Linefeed reader
   *
   * If a given LF byte isn't as expected an exception is thrown
   *
	 * @param fis FileInputStream
   */
  protected void readLinefeed(FileInputStream fis)
    throws Exception
  {
    if (debug) System.out.println("+readLinefeed");
	  FileChannel fc = fis.getChannel();
	  byte[] b = readBytes(fis,0,1,0,null);
    if (b[0] == CR) {
      if (debug) System.out.println("line.separator windows style");
      byte[] b1 = readBytes(fis,0,1,0,null);
      if (b[0] != CR && b1[0] != LF)
        throw new Exception("Carrige return and Linefeed at position "+fc.position()+" not present integer found "+b[0]+" "+b1[0]);
    } else {
      if (b[0] != LF) {
        if (debug) System.out.println("line.separator unix style");
        throw new Exception("Linefeed at position "+fc.position()+" not present integer found "+b[0]);
      }
    }
    //message.addElement(b);
    if (debug) System.out.println("-readLinefeed");
  }

  public void resetMessageBuffer() {
		message.clear();
  }
  
	// Individual methods
  abstract void readFromFile(FileInputStream fis,FileOutputStream fos) throws Exception;
  public void writeToFile(FileOutputStream fos)
		throws Exception
	{
		//if (debug) System.out.println("+writeToFile elements="+message.size());
  	for (int i=0; i<message.size();i++) {
  	  byte[] b = new byte[1];
  	  if (message.elementAt(i) instanceof byte[])
			  b = ((byte[])message.elementAt(i));
  	  else if (message.elementAt(i) instanceof Byte)
  	    b[0] = ((Byte)message.elementAt(i)).byteValue();
  	  else
  	    throw new Exception("Unhandled object attempted to be written! "+(message.elementAt(i)));
			//if (debug) System.out.println(new String(b,"UTF-8"));
	    fos.write(b);
		  fos.flush();
		}
		//if (debug) System.out.println("-writeToFile");
  }

  /**
   * This ia a helper method for fixed length structures. It maps from
   * csv to xml
   * 
   * @param srcPosition csv start position
   * @param fieldLength field byte length
   * @return byte[] result bytes denoting the requestet field
   */
  protected byte[] getFieldValue(int srcPosition, int fieldLength) 
    throws Exception
  {
    byte[] _bytes =  new byte[fieldLength];
    System.arraycopy(payload,srcPosition-1,_bytes,0,_bytes.length);    
    return _bytes;
  }

  /**
   * This ia a helper method for fixed length structures. It maps from
   * csv to xml
   * 
   * @param srcPosition csv start position
   * @param fieldLength field byte length
   * @param fieldPosition xml position in message Vector
   */
  protected void setFieldValue(int srcPosition, int fieldLength, int fieldPosition) 
    throws Exception
  {
    byte[] _bytes =  new byte[fieldLength];
    System.arraycopy(payload,srcPosition-1,_bytes,0,_bytes.length);        
    message.setElementAt(new String(_bytes,"ISO8859_1").getBytes("UTF-8"),fieldPosition);
  }
}