package dk.skat.akfa.transformation.javacallout;

import java.util.zip.GZIPOutputStream;
import java.util.Date;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.nio.channels.FileChannel;

import java.net.URL;
import java.net.URLClassLoader;
import java.io.UnsupportedEncodingException;

/**
 * <p>This POJO is intended to be in the portal and the application server.</p>
 *
 * <p>A message is uploaded in the portal and stored on disk. The fixed csv format is given to this class
 * either as a absolute filename or the csv string to parse. A new file pointer or string is returned.
 *
 * The result is gziped and base 64 encoded SKAT xml.</p>
 *
 * Subversion http://versionhost.ccta.dk/svn/SKAT/products/branches/Fase1/eService/AKFATransformer
 *
 * @author Rene Hjortskov Nielsen, SKAT
 * @version 0.1
 */
public class csv2xml
{
  /**
   * <p>Transform takes a filename to a file containing csv data
   * and transform them to a base64 encoded gzip containing the message
   * as SKAT xml.</p>
   *
   * <p>It is required that we work on parts of input at all times to lower
   * memory consumption. Where applicaple file streams will be used. We also
   * have to take special care for non-UTF-8 characters, which have special
   * meaning to DK AKFA because it can represent a delimiter in the csv data
   * record.
   * </p>
   *
   * @param body64str String containing the payload to be transformed
   * @param datadir Pathlocation for temporary streams
   * @param debug Enable POJO in debug mode for extra logging
   * @return String BASE64 encoded String with an gzipped xml payload
   */
  public static String transform(String body64str, String datadir, boolean debug) {
	 	try {
	 		return new String(transform(body64str.getBytes("UTF-8"), datadir, debug),"UTF-8");
	 	} catch (Exception e) {
	 		throw new IllegalArgumentException(e);
	 	}
  }

  /**
   * <p>Transform takes a filename to a file containing csv data
   * and transform them to a base64 encoded gzip containing the message
   * as SKAT xml.</p>
   *
   * <p>It is required that we work on parts of input at all times to lower
   * memory consumption. Where applicaple file streams will be used. We also
   * have to take special care for non-UTF-8 characters, which have special
   * meaning to DK AKFA because it can represent a delimiter in the csv data
   * record.
   * </p>
   *
   * @param body64 Byte array containing the payload to be transformed
   * @param datadir Pathlocation for temporary streams
   * @param debug Enable POJO in debug mode for extra logging
   * @return byte[] BASE64 encoded byte array with an gzipped xml payload
   */
  public static byte[] transform(byte[] body64, String datadir, boolean debug) {
    return null;
  }

  /**
   * <p>Transform takes a filename to a file containing csv data
   * and transform them to a base64 encoded gzip containing the message
   * as SKAT xml.</p>
   *
   * <p>It is required that we work on parts of input at all times to lower
   * memory consumption. Where applicaple file streams will be used.
   * </p>
   *
   * <p>The output is inserted into the <base64> element before its sent on
   * the wire. The wire xml differs from the original AKFA service with regard
   * to the xml wrapping the base64 encoded data. This is, however, unwrapped
   * at the receiving end.
   *
   * The receiving part in the AKFA system is a tactical service, which means that
   * the unwrapping has to map to the tactical interface and not the fase1 modelled
   * interface.
   * </p>
   *
   * @param filename_csv String containing a filename pointing to the payload to be transformed
   * @param datadir Pathlocation for temporary streams
   * @param debug Enable POJO in debug mode for extra logging
   * @param afgiver Provided by the caller, i.e. the portal
   * @param indsender Provided by the caller, i.e. the portal
   * @return String filename BASE64 encoded gzipped xml payload
   */
  public static String transformfile(String filename_csv, String datadir, boolean debug, String afgiver, String indsender) {
	  ByteArrayOutputStream byteos = null;
	  FileInputStream fis;
	  FileOutputStream fos;
		long startTime = 0L;
		long endTime = 0L;
		startTime = (new Date()).getTime();

    // Filename needs to be thread safe, hence add milliseconds to filename
    String filename = datadir + System.currentTimeMillis();
    try {
      Thread tmpThread = new Thread();
      filename = datadir + "csv2xml_" + tmpThread.getName() + "_" + System.currentTimeMillis();
      if (debug) System.out.println("Using prefix for file: "+ filename);
    } catch (Exception e) {
      // NOP
    }

		String filename_base64 = filename+".xml.gz.base64";
		String filename_zip = filename+".xml.gz";
		String filename_skatxml = filename+".xml";
		try {
	    // loop while data in body
	    fos = new FileOutputStream(filename_skatxml);
	    File f = new File(filename_csv);
	    long flen = f.length();
	    fis = new FileInputStream(f);
	    FileChannel fc = fis.getChannel();
      int linelength = 232;
      int rowCount = (int)flen / linelength; // Fixed fileformat 232 chars per line when windows formattet, otherwise 231
	    if (debug) {
        System.out.println("File size: " + flen);
        System.out.println("File size % 232: " + (flen%232L) + ", individ count = "+(flen/232L));
        System.out.println("File size % 231: " + (flen%231L) + ", individ count = "+(flen/231L));
      }            
      
      // Must read footer 
      fc.position(((long)rowCount-1)*linelength);
	    if (debug) System.out.println("Free memory step 3a: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" Analyse SLUTINDIVID, Rows "+rowCount);
      AKFAXMLMessageFooter xmlfooter = null;
      try {
        xmlfooter = new AKFAXMLMessageFooter(debug);
        xmlfooter.readFromFile(fis,fos);
      } catch (Exception footer) {
        linelength=231;
        rowCount = (int)flen / linelength; // Fixed fileformat 232 chars per line when windows formattet, otherwise 231
        fc.position(((long)rowCount-1)*linelength);
        xmlfooter = new AKFAXMLMessageFooter(debug);
        xmlfooter.readFromFile(fis,fos);
      }
      // This situation is occuring when rowCount%231=0 == rowCount%232=0 
      if ((flen%232L)==0 && (flen%231L)==0) {      
        rowCount = (int)xmlfooter.getIndivider() + 2;
        linelength=(int)(flen/rowCount);
        if (linelength!=231 && linelength!=232)
          throw new Exception("Fixed line length exceeded 230 + [CR] + LF : "+linelength);
        rowCount = (int)flen / linelength; // Fixed fileformat 232 chars per line when windows formattet, otherwise 231
        fc.position(((long)rowCount-1)*linelength);
        xmlfooter = new AKFAXMLMessageFooter(debug);
        xmlfooter.readFromFile(fis,fos);
      }      
	    fc.position(0L);

	    if (debug) System.out.println("Free memory step 1: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" Analyse STARTINDIVID, Rows "+rowCount);
	    AKFAXMLMessageHeader xmlheader = new AKFAXMLMessageHeader(debug, xmlfooter.getBrugernummer(), rowCount, afgiver, indsender);
	    xmlheader.readFromFile(fis,fos);
	    xmlheader.writeToFile(fos);
      xmlheader.resetMessageBuffer();
	    rowCount--;
	    xmlheader = null;

	    AKFAXMLMessageInterface xmlmessage = (AKFAXMLMessageInterface)new AKFAXMLMessage(debug);
  	  if (xmlmessage==null)
  	    throw new Exception("Message parser failed to instantiate");
	    while (rowCount>1) {
  	    if (debug) System.out.println("Free memory step 2: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" Analyse INDIVID, Rows "+rowCount);
		    xmlmessage.readFromFile(fis,fos);
  		  xmlmessage.writeToFile(fos);
	      rowCount--;
  	    if ((fc.position()%linelength)!=0)
  	      throw new Exception("Unexpected position in line "+fc.position()+" of "+flen+" at row "+rowCount);
        xmlmessage.resetMessageBuffer();
	    }
      xmlmessage.resetMessageBuffer();
		  xmlmessage = null;

	    if (debug) System.out.println("Free memory step 3b: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" Analyse SLUTINDIVID, Rows "+rowCount);
		  xmlfooter.writeToFile(fos);
	    rowCount--;
	    xmlfooter = null;
	    if (rowCount!=0)
	      throw new Exception("Unexpected lines left "+rowCount);
	    fc = fis.getChannel();
	    //if (fc.position()<flen) {
	    if (((int)fc.position()%linelength)!=0) {
	      long pos = fc.position();
	      if (debug) {
	        System.out.println("Bytes left: "+(int)(flen-fc.position()));
	        byte[] b = new byte[(int)(flen-fc.position())];
	        fis.read(b);
	        for (int i=0; i<b.length; i++)
	          System.out.print(i+"="+b[i]+" ");
	        System.out.println("");
	      }
	      throw new Exception("Unexpected position in line "+(((int)flen)/linelength)+" at "+pos+" of "+flen);
	    }
	    fis.close();
	    fos.close();

	    if (debug) System.out.println("Free memory step 4: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" gzip the outputted stream, Rows "+rowCount);
	    fis = new FileInputStream(filename_skatxml);
	    fos = new FileOutputStream(filename_zip);
	    GZIPOutputStream zip = new GZIPOutputStream(fos);
	    int i = 0;
	    byte[] buf = new byte[1024*1024];
      int len;
      while ((len = fis.read(buf)) > 0) {
        zip.write(buf, 0, len);
      }
	    fis.close();
	    zip.finish();
	    fos.close();

	    if (debug) System.out.println("Free memory step 5: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" base64 encode the gzip archieve");
	    f = new File(filename_zip);
	    flen = f.length();
	    fis = new FileInputStream(f);
	    byteos = new ByteArrayOutputStream((int)flen);
      Base64 base64 = new Base64();
      base64.encode(fis, byteos);
	    fis.close();

//+++ debug
			// ??? if (debug) {
				fos = new FileOutputStream(filename_base64);
				fos.write(byteos.toByteArray());
				fos.flush();
				fos.close();
			//}
//---

  		if (debug) {
  		  endTime = (new Date()).getTime();
  		  System.out.println("Elasped ms "+(endTime-startTime));
  		}

  		// In production we remove temporary files
  	  if (!debug) {
  	  	removeFile(filename_csv);
  	  	removeFile(filename_skatxml);
  	  	//removeFile(filename_base64);
  	  	removeFile(filename_zip);
  	  }
		} catch(Exception e) {
		  String errstr = "Error in message ["+filename_csv+"]:"+e+"\n";
		  StackTraceElement[] ste = e.getStackTrace();
		  for (int i=0; i<ste.length; i++) {
		    errstr += ste[i].toString()+"\n";
		  }
			byteos = null;
	 	  //System.out.println(""+errstr);
      throw new IllegalArgumentException(errstr);
 		}

    return filename_base64;// new String(byteos.toByteArray());
  }

  private static void removeFile(String filename) {
    try {
      File f =new File(filename);
      if (f.exists())
        f.delete();
    } catch (Exception e) {
      // NOP
    }
  }

  private static void printClasspath()
  {
    try {
      //Get the System Classloader
      ClassLoader sysClassLoader = ClassLoader.getSystemClassLoader();
      //Get the URLs
      URL[] urls = ((URLClassLoader)sysClassLoader).getURLs();
      for(int i=0; i< urls.length; i++) {
        System.out.println(urls[i].getFile());
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

}