package dk.skat.akfa.transformation.javacallout;

import java.util.zip.GZIPInputStream;
import java.util.Date;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.nio.channels.FileChannel;
import java.io.UnsupportedEncodingException;

/**
 * <p>This POJO is intended to be inserted into a ALSB message flow.</p>
 *
 * <p>A base64 encoded gz message wrapping SKAT XML has to be decoded. A new file pointer or string is returned.
 *
 * The result is SKAT xml.</p>
 *
 * @author Ren Hjortskov Nielsen, SKAT
 * @version 0.1
 */
public class xml2csv
{

  /**
   * <p>Transform takes a base64 encoded gzipped payload containing SKAT xml
   * messages and transform them to a base64 encoded message with the AKFA csv.</p>
   *
   * <p>It is required that we work on parts of input at all times to lower
   * memory consumption. Where applicaple file streams will be used. We also
   * have to take special care for non-UTF-8 characters, which have special
   * meaning to DK AKFA because it can represent a delimiter in the csv data
   * record.
   * </p>
   *
   * @param body64str Payload to be transformed as a String
   * @param datadir Pathlocation for temporary stream e.g. "/tmp/"
   * @param debug Enable POJO in debug mode for extra logging
   * messages. When disabled, temporary files are deleted.
   * @return String Decoded SKAT XML file pointer
   */
  public static String transform(String body64str, String datadir, boolean debug) {
	 	try {
	 		return transform(body64str.getBytes("UTF-8"), datadir, debug);
	 	} catch (Exception e) {
	 		throw new IllegalArgumentException(e);
	 	}
  }

  /**
   * <p>Transform takes a base64 encoded gzipped payload containing SKAT xml
   * messages and transform them to a base64 encoded message with the AKFA csv.</p>
   *
   * <p>It is required that we work on parts of input at all times to lower
   * memory consumption. Where applicaple file streams will be used. We also
   * have to take special care for non-UTF-8 characters, which have special
   * meaning to DK AKFA because it can represent a delimiter in the csv data
   * record.
   * </p>
   *
   * @param body64 Byte array containing the payload to be transformed
   * @param datadir Pathlocation for temporary stream
   * @param debug Enable POJO in debug mode for extra logging
   * messages. When disabled, temporary files are deleted.
   * @return String Decoded SKAT XML file pointer
   */
  public static String transform(byte[] body64, String datadir, boolean debug) {
    // Filename needs to be thread safe, hence add millseconds to filename
    String filename = datadir + System.currentTimeMillis();
    try {
      Thread tmpThread = new Thread();
      filename = "xml2csv_" + tmpThread.getName() + "_" + System.currentTimeMillis();
      if (debug) System.out.println("Using prefix for file: "+ filename);
    } catch (Exception e) {
      // NOP
    }
		String filename_byte = filename+".xml.base64.gz.byte";
	 	try {
  	  FileOutputStream fos = new FileOutputStream(datadir + filename_byte);
  	  fos.write(body64);
  	  fos.close();
	 		return transformfile(filename_byte, datadir, debug);
	 	} catch (Exception e) {
	 		throw new IllegalArgumentException(e);
	 	} finally {
  	  if (!debug) {
  	  	removeFile(filename_byte);
  	  }
	 	}
  }

  /**
   * <p>Transform takes a filename to a file containing base64 encoded gz data
   * and decodes them to the containing message as SKAT xml.</p>
   *
   * @param filename_base64 String containing a filename pointing to the payload to be transformed
   * @param datadir Pathlocation for temporary streams
   * @param debug Enable POJO in debug mode for extra logging
   * @return String Decoded SKAT XML file pointer
   */
  public static String transformfile(String filename_base64, String datadir, boolean debug) {
    boolean escapeSequenceFound = false;
	  byte[] byteos = new byte[0];
	  FileInputStream fis;
	  FileOutputStream fos;
		long startTime = 0L;
		long endTime = 0L;
  	startTime = (new Date()).getTime();

    // Filename needs to be thread safe, hence add millseconds to filename
    String filename = datadir + System.currentTimeMillis();
    try {
      Thread tmpThread = new Thread();
      filename = datadir + "xml2csv_" + tmpThread.getName() + "_" + System.currentTimeMillis();
      if (debug) System.out.println("Using prefix for file: "+ filename);
    } catch (Exception e) {
      // NOP
    }

		String filename_zip = filename+".xml.gz";
		String filename_skatxml = filename+".xml";
		String filename_csv = filename+".csv";

		try {
	    if (debug) System.out.println("Free memory step 1: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" Decode BASE64");
	    // Step 1 - Decode BASE64
	    fos = new FileOutputStream(filename_zip);
	    fis = new FileInputStream(datadir + filename_base64);
      Base64 base64 = new Base64();
      base64.decode(fis, fos);
	    fos.close();
	    fis.close();

	    if (debug) System.out.println("Free memory step 2: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" Gunzip body");
	    // Step 2 - Unzip gzipped body.
	    GZIPInputStream zip = new GZIPInputStream(new FileInputStream(filename_zip),1024*1024);
	    fos = new FileOutputStream(filename_skatxml);
	    byte[] buf = new byte[1024*1024];
      int len;
      while ((len = zip.read(buf)) > 0) {
        fos.write(buf, 0, len);
      }
      buf = null;
	    zip.close();
	    fos.close();

	    if (debug) System.out.println("Free memory step 3: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" Process SKAT xml using a SAX parser with xlst");
			// Step 3 - Process SKAT xml using a SAX parser with xlst
	    File f = new File(filename_skatxml);
	    long flen = f.length();
	    fis = new FileInputStream(f);
	    fos = new FileOutputStream(filename_csv);
			XMLAKFAMessageParser xmlparser = new XMLAKFAMessageParser(debug,fis,fos,flen);
			escapeSequenceFound = xmlparser.parse();
			fis.close();
			fos.close();
			xmlparser=null;

/*
	    if (debug) System.out.println("Free memory step 4a: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" base64 encode the AKFA csv");
	    // Step 6 - base64 encode AKFA csv
      f = new File(filename_eucsv2);
      fis = new FileInputStream(f);
	    fos = new FileOutputStream(filename_eucsv_base64);
      base64.encode(fis, fos);
	    fis.close();
	    fos.close();
	    base64 = null;

      if (debug) System.out.println("Free memory step 4b: "+Runtime.getRuntime().freeMemory()+" of "+Runtime.getRuntime().totalMemory()+" base64 encode the AKFA csv");
	    f = new File(filename_eucsv_base64);
	    flen = f.length();
	    fis = new FileInputStream(f);
      // Use body64 instead, could safe some memory, but also impose a small risk in case of a reading error
	    byteos = new byte[(int)flen];
	    fis.read(byteos,0,(int)flen);
      fis.close();
*/
  		if (debug) {
  		  endTime = (new Date()).getTime();
  		  System.out.println("Elasped ms "+(endTime-startTime));
  		}

  		// In production we remove temporary files
  	  if (!debug) {
  	  	removeFile(filename_zip);
  	  	removeFile(filename_skatxml);
  	  }

		} catch(Exception e) {
		  String errstr = "Error in message ["+filename_base64+"]:\n"+e+"\n";
		  StackTraceElement[] ste = e.getStackTrace();
		  for (int i=0; i<ste.length; i++) {
		    errstr += ste[i].toString()+"\n";
		  }
			byteos = null;
 	 	  //System.out.println(""+errstr);
      throw new IllegalArgumentException(errstr);
 		}

    return filename_csv;
  }

  private static void removeFile(String filename) {
    try {
      File f =new File(filename);
      if (f.exists())
        f.delete();
    } catch (Exception e) {
      // NOP
    }
  }
}